"""
The 'worktoy.work_test' package provides a subclass of 'unittest.TestCase'.
"""
#  AGPL-3.0 license
#  Copyright (c) 2026 Asger Jon Vistisen
from __future__ import annotations

from ._base_test import BaseTest

__all__ = [
  'BaseTest',
  ]
